import logging

import eyepy as ep
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtWidgets import QWidget

from eyelab.models.treeitemdelegate import TreeItemDelegate
from eyelab.models.treeview.itemmodel import (
    EnfaceTreeItemModel,
    TreeItem,
    TreeItemModel,
    VolumeTreeItemModel,
)
from eyelab.tools import area_tools, basic_tools, line_tools
from eyelab.views.ui.ui_scene_tab import Ui_SceneTab

logger = logging.getLogger("eyelab.viewtab")


class ViewTab(QWidget, Ui_SceneTab):
    def __init__(self, data, parent=None):
        super().__init__(parent)
        self.setupUi(self)
        self.data = data
        self.model: TreeItemModel = None

        self.line_tools = line_tools()
        self.area_tools = area_tools()
        self.basic_tools = basic_tools()

        self.upButton.clicked.connect(self.layer_up)
        self.downButton.clicked.connect(self.layer_down)
        self.downButton.clicked.connect(self.layer_down)
        self.opacitySlider.valueChanged.connect(self.set_opacity)

        # Todo: Fix bugs to enable featuress
        self.opacitySliderLabel.hide()
        self.opacitySlider.hide()
        self.upButton.hide()
        self.downButton.hide()

        self.current_tool = None
        self.setTools(self.basic_tools)

    # @property
    # def workspace(self):
    #    return self.parent()

    def set_model(self, model: TreeItemModel):
        self.model = model
        self.addButton.clicked.connect(self._add_annotation)
        self.deleteButton.clicked.connect(self._remove_annotation)

    def _remove_annotation(self):
        index = self.imageTreeView.selectionModel().currentIndex()
        self.model.remove_annotation(index)

    def _add_annotation(self):
        self.model.add_annotation()

    def configure_imageTreeView(self):
        self.imageTreeView.setModel(self.model)
        self.imageTreeView.header().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        for col in range(1, self.model.columnCount()):
            self.imageTreeView.hideColumn(col)
        self.imageTreeView.setHeaderHidden(True)
        self.imageTreeView.setItemDelegate(TreeItemDelegate(self.imageTreeView))
        self.imageTreeView.setRootIsDecorated(False)

        self.imageTreeView.selectionModel().currentRowChanged.connect(
            self.on_currentChanged
        )

        self.imageTreeView.expandAll()

        self.imageTreeView.setContextMenuPolicy(QtCore.Qt.ActionsContextMenu)
        duplicate_action = QtGui.QAction(self)
        duplicate_action.setText("Duplicate All")
        duplicate_action.triggered.connect(self.duplicate_volume)

        self.imageTreeView.addAction(duplicate_action)

    def duplicate_volume(self):
        self.model.duplicate_volume(self.imageTreeView.selectionModel().currentIndex())

    def tools(self):
        return self._tools

    def setTools(self, tools=None, default="inspection"):
        if tools is None:
            tools = self.basic_tools
        self._tools = tools
        self._set_tool_buttons()
        self._switch_to_tool(default)()

    def _set_tool_buttons(self):
        layout = self.toolsWidget.layout()
        # Remove previous tools
        for i in reversed(range(layout.count())):
            layout.itemAt(i).widget().setParent(None)
        # Add current tools
        for i, (name, tool) in enumerate(self.tools().items()):
            layout.addWidget(tool.button, 1, i, 1, 1)
            tool.button.clicked.connect(self._switch_to_tool(name))

    def _switch_to_tool(self, name):
        def func():
            tool = self._tools[name]
            for n, t in self._tools.items():
                t.button.setChecked(False)
                t.disable()
            tool.button.setChecked(True)

            # Show tool options
            self.toolboxWidget.layout().removeWidget(self.optionsWidget)
            self.optionsWidget.setParent(None)
            self.toolboxWidget.layout().addWidget(tool.options_widget)

            self.optionsWidget = tool.options_widget
            self.toolboxWidget.repaint()

            self.current_tool = tool
            tool.enable()

            # if self.tool.paint_preview.scene() == self.scene():
            #    self.scene().removeItem(self.tool.paint_preview)
            # self.tool.paint_preview.setParentItem(
            #    self.scene().activePanel())

        return func

    def set_opacity(self, value):
        self.model.tree_root.setOpacity(value / 100)

    def layer_up(self):
        selected = self.imageTreeView.selectionModel().currentIndex()

        parent = selected.parent()
        row1 = selected.row()
        row2 = row1 - 1

        if 0 < row1 <= parent.internalPointer().childCount():
            self.model.switchRows(row2, row1, parent)
            # self.imageTreeView.selectionModel().currentRowChanged.emit(
            #    self.model.index(row1, 0, parent),
            #    self.model.index(row2, 0, parent))

    def layer_down(self):
        selected = self.imageTreeView.selectionModel().currentIndex()

        parent = selected.parent()
        source_row = selected.row()
        target_row = source_row + 1

        if 0 <= source_row < parent.internalPointer().childCount() - 1:
            self.model.switchRows(source_row, target_row, parent)
            # self.imageTreeView.selectionModel().currentRowChanged.emit(
            #    self.model.index(row1, 0, parent),
            #    self.model.index(row2, 0, parent))

    @QtCore.Slot("QModelIndex", "QModelIndex")
    def on_currentChanged(self, current, previous):
        current_tree_item = self.model.getItem(current)
        previous_tree_item = self.model.getItem(previous)

        # Change activated item
        self.model.activate(current)
        self.model.deactivate(previous)

        # Set tools for current item
        if current_tree_item.parent.data("name") == "Layers":
            self.line_tools["spline"].options_widget.set_data(
                current_tree_item.itemData
            )
            self.setTools(self.line_tools, default="spline")
        elif current_tree_item.parent.data("name") == "Areas":
            self.setTools(self.area_tools, default="pen")


class VolumeTab(ViewTab):
    def __init__(self, data: ep.EyeVolume, parent=None):
        super().__init__(data, parent)
        self.set_model(VolumeTreeItemModel(data=data, parent=self))
        self.configure_imageTreeView()

    def next_slice(self):
        self.model.next_slice(self.current_tool)
        self.model.scene.addItem(self.current_tool.paint_preview)

    def last_slice(self):
        self.model.last_slice(self.current_tool)
        self.model.scene.addItem(self.current_tool.paint_preview)

    @property
    def current_item(self):
        model_index = self.imageTreeView.selectionModel().currentIndex()
        tree_item = self.model.getItem(model_index)
        if type(tree_item) == TreeItem:
            item = self.model.annotation_items[id(tree_item.annotation)][
                self.model.current_slice
            ]
            return item


class EnfaceTab(ViewTab):
    def __init__(self, data: ep.EyeEnface, parent=None):
        super().__init__(data, parent)
        self.set_model(EnfaceTreeItemModel(data=data, parent=self))

        self.configure_imageTreeView()

    @property
    def current_item(self):
        model_index = self.imageTreeView.selectionModel().currentIndex()
        tree_item = self.model.getItem(model_index)
        if type(tree_item) == TreeItem:
            item = self.model.annotation_items[id(tree_item.annotation)]
            return item



from PySide6 import QtWidgets

from eyelab.views.ui.ui_text_window import Ui_TextWindow


class HelpWindow(QtWidgets.QDialog, Ui_TextWindow):
    def __init__(self, md_str, parent=None):
        super().__init__(parent)
        self.setupUi(self)
        self.textEdit.setMarkdown(md_str)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_main_window.ui'
##
## Created by: Qt User Interface Compiler version 6.1.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *  # type: ignore
from PySide6.QtGui import *  # type: ignore
from PySide6.QtWidgets import *  # type: ignore

from . import resources_rc


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(836, 618)
        self.action_undo = QAction(MainWindow)
        self.action_undo.setObjectName(u"action_undo")
        self.action_undo.setCheckable(False)
        self.action_undo.setEnabled(True)
        icon = QIcon()
        icon.addFile(
            u":/icons/icons/baseline-undo-24px.svg", QSize(), QIcon.Normal, QIcon.Off
        )
        self.action_undo.setIcon(icon)
        self.action_redo = QAction(MainWindow)
        self.action_redo.setObjectName(u"action_redo")
        icon1 = QIcon()
        icon1.addFile(
            u":/icons/icons/baseline-redo-24px.svg", QSize(), QIcon.Normal, QIcon.Off
        )
        self.action_redo.setIcon(icon1)
        self.action_import_vol = QAction(MainWindow)
        self.action_import_vol.setObjectName(u"action_import_vol")
        self.action_save_annotations = QAction(MainWindow)
        self.action_save_annotations.setObjectName(u"action_save_annotations")
        self.action_save_annotations.setEnabled(True)
        self.action_save_annotations.setVisible(True)
        self.action_import_bsfolder = QAction(MainWindow)
        self.action_import_bsfolder.setObjectName(u"action_import_bsfolder")
        self.action_layer_annotation_guide = QAction(MainWindow)
        self.action_layer_annotation_guide.setObjectName(
            u"action_layer_annotation_guide"
        )
        self.action_area_annotation_guide = QAction(MainWindow)
        self.action_area_annotation_guide.setObjectName(u"action_area_annotation_guide")
        self.action_registration_guide = QAction(MainWindow)
        self.action_registration_guide.setObjectName(u"action_registration_guide")
        self.action_registration_guide.setVisible(False)
        self.action_shortcut_sheet = QAction(MainWindow)
        self.action_shortcut_sheet.setObjectName(u"action_shortcut_sheet")
        self.action_introduction = QAction(MainWindow)
        self.action_introduction.setObjectName(u"action_introduction")
        self.action_import_retouch = QAction(MainWindow)
        self.action_import_retouch.setObjectName(u"action_import_retouch")
        self.action_import_duke = QAction(MainWindow)
        self.action_import_duke.setObjectName(u"action_import_duke")
        self.action_import_hexml = QAction(MainWindow)
        self.action_import_hexml.setObjectName(u"action_import_hexml")
        self.action_load_annotations = QAction(MainWindow)
        self.action_load_annotations.setObjectName(u"action_load_annotations")
        self.action_load_annotations.setEnabled(True)
        self.action_load_annotations.setVisible(True)
        self.action_save_annotations_as = QAction(MainWindow)
        self.action_save_annotations_as.setObjectName(u"action_save_annotations_as")
        self.action_save_annotations_as.setEnabled(True)
        self.action_save_annotations_as.setVisible(True)
        self.action_open = QAction(MainWindow)
        self.action_open.setObjectName(u"action_open")
        self.action_save = QAction(MainWindow)
        self.action_save.setObjectName(u"action_save")
        self.action_save_as = QAction(MainWindow)
        self.action_save_as.setObjectName(u"action_save_as")
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout = QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(MainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 836, 22))
        self.menuFile = QMenu(self.menubar)
        self.menuFile.setObjectName(u"menuFile")
        self.menuUpload = QMenu(self.menuFile)
        self.menuUpload.setObjectName(u"menuUpload")
        self.menuAnnotations = QMenu(self.menuFile)
        self.menuAnnotations.setObjectName(u"menuAnnotations")
        self.menuAnnotations.setEnabled(False)
        self.menuEdit = QMenu(self.menubar)
        self.menuEdit.setObjectName(u"menuEdit")
        self.menuEdit.setEnabled(True)
        self.menuEdit.setTearOffEnabled(False)
        self.menuHelp = QMenu(self.menubar)
        self.menuHelp.setObjectName(u"menuHelp")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuEdit.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.action_open)
        self.menuFile.addAction(self.action_save)
        self.menuFile.addAction(self.action_save_as)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.menuUpload.menuAction())
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.menuAnnotations.menuAction())
        self.menuUpload.addAction(self.action_import_vol)
        self.menuUpload.addAction(self.action_import_hexml)
        self.menuUpload.addSeparator()
        self.menuUpload.addAction(self.action_import_retouch)
        self.menuUpload.addAction(self.action_import_duke)
        self.menuUpload.addSeparator()
        self.menuUpload.addAction(self.action_import_bsfolder)
        self.menuAnnotations.addAction(self.action_load_annotations)
        self.menuAnnotations.addSeparator()
        self.menuAnnotations.addAction(self.action_save_annotations)
        self.menuAnnotations.addAction(self.action_save_annotations_as)
        self.menuHelp.addAction(self.action_introduction)
        self.menuHelp.addAction(self.action_layer_annotation_guide)
        self.menuHelp.addAction(self.action_area_annotation_guide)
        self.menuHelp.addAction(self.action_registration_guide)
        self.menuHelp.addSeparator()
        self.menuHelp.addAction(self.action_shortcut_sheet)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)

    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(
            QCoreApplication.translate("MainWindow", u"EyeLab", None)
        )
        self.action_undo.setText(
            QCoreApplication.translate("MainWindow", u"&Undo", None)
        )
        # if QT_CONFIG(shortcut)
        self.action_undo.setShortcut(
            QCoreApplication.translate("MainWindow", u"Ctrl+Z", None)
        )
        # endif // QT_CONFIG(shortcut)
        self.action_redo.setText(
            QCoreApplication.translate("MainWindow", u"&Redo", None)
        )
        # if QT_CONFIG(shortcut)
        self.action_redo.setShortcut(
            QCoreApplication.translate("MainWindow", u"Ctrl+Y", None)
        )
        # endif // QT_CONFIG(shortcut)
        self.action_import_vol.setText(
            QCoreApplication.translate("MainWindow", u"Heyex Raw (.vol)", None)
        )
        self.action_save_annotations.setText(
            QCoreApplication.translate("MainWindow", u"Save", None)
        )
        self.action_import_bsfolder.setText(
            QCoreApplication.translate("MainWindow", u"B-Scans from folder", None)
        )
        self.action_layer_annotation_guide.setText(
            QCoreApplication.translate("MainWindow", u"Layer Annotation Guide", None)
        )
        self.action_area_annotation_guide.setText(
            QCoreApplication.translate("MainWindow", u"Area Annotation Guide", None)
        )
        self.action_registration_guide.setText(
            QCoreApplication.translate("MainWindow", u"Registration Guide", None)
        )
        self.action_shortcut_sheet.setText(
            QCoreApplication.translate("MainWindow", u"Shortcut Sheet", None)
        )
        self.action_introduction.setText(
            QCoreApplication.translate("MainWindow", u"Introduction", None)
        )
        self.action_import_retouch.setText(
            QCoreApplication.translate("MainWindow", u"RETOUCH", None)
        )
        self.action_import_duke.setText(
            QCoreApplication.translate("MainWindow", u"Duke (Farsiu 2014)", None)
        )
        self.action_import_hexml.setText(
            QCoreApplication.translate("MainWindow", u"Heyex XML (.xml)", None)
        )
        self.action_load_annotations.setText(
            QCoreApplication.translate("MainWindow", u"Load", None)
        )
        self.action_save_annotations_as.setText(
            QCoreApplication.translate("MainWindow", u"Save As", None)
        )
        self.action_open.setText(
            QCoreApplication.translate("MainWindow", u"&Open", None)
        )
        # if QT_CONFIG(shortcut)
        self.action_open.setShortcut(
            QCoreApplication.translate("MainWindow", u"Ctrl+O", None)
        )
        # endif // QT_CONFIG(shortcut)
        self.action_save.setText(
            QCoreApplication.translate("MainWindow", u"&Save", None)
        )
        # if QT_CONFIG(shortcut)
        self.action_save.setShortcut(
            QCoreApplication.translate("MainWindow", u"Ctrl+S", None)
        )
        # endif // QT_CONFIG(shortcut)
        self.action_save_as.setText(
            QCoreApplication.translate("MainWindow", u"Save &As", None)
        )
        # if QT_CONFIG(shortcut)
        self.action_save_as.setShortcut(
            QCoreApplication.translate("MainWindow", u"Ctrl+A", None)
        )
        # endif // QT_CONFIG(shortcut)
        self.menuFile.setTitle(QCoreApplication.translate("MainWindow", u"&File", None))
        self.menuUpload.setTitle(
            QCoreApplication.translate("MainWindow", u"Import", None)
        )
        self.menuAnnotations.setTitle(
            QCoreApplication.translate("MainWindow", u"Annotations", None)
        )
        self.menuEdit.setTitle(QCoreApplication.translate("MainWindow", u"&Edit", None))
        self.menuHelp.setTitle(QCoreApplication.translate("MainWindow", u"Help", None))

    # retranslateUi



import eyepy as ep
import numpy as np
import qimage2ndarray
from PySide6 import QtCore, QtGui, QtWidgets

from eyelab.models.treeview.itemgroup import ItemGroup


class AreaItem(QtWidgets.QGraphicsPixmapItem):
    def __init__(
        self,
        data: ep.EyeVolumeVoxelAnnotation,
        index: int = None,
        parent: ItemGroup = None,
    ):
        super().__init__(parent=parent)
        self.annotation_data = data
        self.index = index
        if self.index is None:
            self.slice = self.annotation_data.data
        else:
            self.slice = self.annotation_data.data[self.index]

        height, width = self.slice.shape
        self.qimage = QtGui.QImage(width, height, QtGui.QImage.Format_ARGB32)
        color = QtGui.QColor()
        color.setNamedColor(f"#{self.annotation_data.meta['current_color']}")
        self.qimage.fill(color)
        self.alpha_array = qimage2ndarray.alpha_view(self.qimage)
        self.setPixmap(QtGui.QPixmap())
        self.set_data()

        self.setFlag(QtWidgets.QGraphicsItem.ItemIsPanel)

    def update(self):
        self.setVisible(self.annotation_data.meta["visible"])
        self.setZValue(self.annotation_data.meta["z_value"])

        color = QtGui.QColor()
        color.setNamedColor(f"#{self.annotation_data.meta['current_color']}")
        qimage2ndarray.rgb_view(self.qimage)[:] = np.array(
            [color.red(), color.green(), color.blue()]
        )
        self.update_pixmap()
        super().update()

    def shape(self) -> QtGui.QPainterPath:
        path = QtGui.QPainterPath()
        path.addRect(QtCore.QRectF(self.qimage.rect()))
        return path

    def update_pixmap(self):
        pixmap = self.pixmap()
        pixmap.convertFromImage(self.qimage)
        self.setPixmap(pixmap)
        self.slice[...] = self.alpha_array

    def set_data(self):
        self.alpha_array[...] = 0.0
        self.alpha_array[self.slice] = 255
        self.update_pixmap()

    def view(self):
        return self.scene().views()[0]

    def mousePressEvent(self, event):
        self.view().tool.mouse_press_handler(self, event)
        event.accept()

    def mouseReleaseEvent(self, event):
        self.view().tool.mouse_release_handler(self, event)
        event.accept()

    def keyPressEvent(self, event):
        self.view().tool.key_press_handler(self, event)
        event.accept()

    def keyReleaseEvent(self, event):
        self.view().tool.key_release_handler(self, event)
        event.accept()

    def mouseMoveEvent(self, event):
        self.view().tool.mouse_move_handler(self, event)
        event.accept()

    def add_pixels(self, pos, mask):
        size_x, size_y = mask.shape
        offset_x = pos.x() - (size_x - 1) / 2
        offset_y = pos.y() - (size_y - 1) / 2

        for ix, iy in np.ndindex(mask.shape):
            if mask[ix, iy]:
                self.alpha_array[int(offset_y + iy), int(offset_x + ix)] = 255.0
                self.slice[int(offset_y + iy), int(offset_x + ix)] = True

        self.update_pixmap()

    def remove_pixels(self, pos, mask):
        size_x, size_y = mask.shape
        offset_x = pos.x() - (size_x - 1) / 2
        offset_y = pos.y() - (size_y - 1) / 2
        for ix, iy in np.ndindex(mask.shape):
            if mask[ix, iy]:
                self.alpha_array[int(offset_y + iy), int(offset_x + ix)] = 0.0
                self.slice[int(offset_y + iy), int(offset_x + ix)] = False

        self.update_pixmap()



class Tool(object):
    def __init__(self):
        self.cursor = None
        self.button = None
        self.hot_key = None
        self.settings_widget = None

    def enable(self, view):
        pass

    def disable(self, view):
        pass

    def mouse_move_handler(self):
        pass

    def mouse_click_handler(self):
        pass



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_pen_options.ui'
##
## Created by: Qt User Interface Compiler version 6.1.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *  # type: ignore
from PySide6.QtGui import *  # type: ignore
from PySide6.QtWidgets import *  # type: ignore


class Ui_penOptions(object):
    def setupUi(self, penOptions):
        if not penOptions.objectName():
            penOptions.setObjectName(u"penOptions")
        penOptions.resize(126, 56)
        self.verticalLayout = QVBoxLayout(penOptions)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(2, 2, 2, 2)
        self.label = QLabel(penOptions)
        self.label.setObjectName(u"label")
        font = QFont()
        font.setPointSize(8)
        font.setBold(True)
        self.label.setFont(font)

        self.verticalLayout.addWidget(self.label)

        self.formLayout = QFormLayout()
        self.formLayout.setObjectName(u"formLayout")
        self.formLayout.setHorizontalSpacing(2)
        self.formLayout.setVerticalSpacing(2)
        self.formLayout.setContentsMargins(5, -1, -1, -1)
        self.sizeLabel = QLabel(penOptions)
        self.sizeLabel.setObjectName(u"sizeLabel")
        font1 = QFont()
        font1.setPointSize(8)
        self.sizeLabel.setFont(font1)

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.sizeLabel)

        self.sizeSlider = QSlider(penOptions)
        self.sizeSlider.setObjectName(u"sizeSlider")
        self.sizeSlider.setMinimum(1)
        self.sizeSlider.setMaximum(100)
        self.sizeSlider.setValue(5)
        self.sizeSlider.setOrientation(Qt.Horizontal)

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.sizeSlider)

        self.verticalLayout.addLayout(self.formLayout)

        self.verticalSpacer = QSpacerItem(
            20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding
        )

        self.verticalLayout.addItem(self.verticalSpacer)

        self.retranslateUi(penOptions)

        QMetaObject.connectSlotsByName(penOptions)

    # setupUi

    def retranslateUi(self, penOptions):
        penOptions.setWindowTitle("")
        self.label.setText(QCoreApplication.translate("penOptions", u"Pen:", None))
        self.sizeLabel.setText(QCoreApplication.translate("penOptions", u"Size:", None))

    # retranslateUi



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_3d_viewer.ui'
##
## Created by: Qt User Interface Compiler version 6.1.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from oat.views.custom import CustomGraphicsView
from PySide6.QtCore import *  # type: ignore
from PySide6.QtGui import *  # type: ignore
from PySide6.QtWidgets import *  # type: ignore


class Ui_Viewer3D(object):
    def setupUi(self, Viewer3D):
        if not Viewer3D.objectName():
            Viewer3D.setObjectName(u"Viewer3D")
        Viewer3D.resize(400, 300)
        self.verticalLayout_2 = QVBoxLayout(Viewer3D)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.graphicsView3D = CustomGraphicsView(Viewer3D)
        self.graphicsView3D.setObjectName(u"graphicsView3D")
        self.graphicsView3D.setMinimumSize(QSize(400, 400))
        self.graphicsView3D.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.graphicsView3D.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self.verticalLayout_2.addWidget(self.graphicsView3D)

        self.spinBox = QSpinBox(Viewer3D)
        self.spinBox.setObjectName(u"spinBox")

        self.verticalLayout_2.addWidget(self.spinBox)

        self.retranslateUi(Viewer3D)

        QMetaObject.connectSlotsByName(Viewer3D)

    # setupUi

    def retranslateUi(self, Viewer3D):
        Viewer3D.setWindowTitle(QCoreApplication.translate("Viewer3D", u"Form", None))

    # retranslateUi



import logging
from collections import namedtuple

import eyepy as ep
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import QRectF, Qt
from PySide6.QtWidgets import QGraphicsSceneContextMenuEvent

from eyelab.models.utils import array2qgraphicspixmapitem

logger = logging.getLogger("eyelab.scene")

Line = namedtuple("Line", ["a", "b", "c"])
Point = namedtuple("Point", ["x", "y"])


class CustomGraphicsScene(QtWidgets.QGraphicsScene):
    def __init__(self, parent, data: ep.EyeBscan, *args, **kwargs):
        super().__init__(*args, **kwargs, parent=parent)
        self.data = data

        self._widthForHeightFactor = 1

        self.background_on = True
        self.fake_cursor = self.addPixmap(
            QtGui.QPixmap(":/cursors/cursors/navigation_cursor.svg")
        )
        self.fake_cursor.setFlag(QtWidgets.QGraphicsItem.ItemIgnoresTransformations)
        self.fake_cursor.hide()

        self.set_image()

        self.setItemIndexMethod(QtWidgets.QGraphicsScene.NoIndex)

        logger.debug("CustomGraphicsScene: __init__ done")

    def drawBackground(self, painter: QtGui.QPainter, rect: QtCore.QRectF):
        if self.background_on:
            painter.fillRect(self.sceneRect(), self.backgroundBrush())
        else:
            painter.fillRect(self.sceneRect(), QtCore.Qt.NoBrush)

    def set_image(self):
        pixmap_item = array2qgraphicspixmapitem(self.data.data)
        pixmap = pixmap_item.pixmap()
        self.shape = (pixmap.height(), pixmap.width())
        self.setSceneRect(QRectF(pixmap.rect()))
        self._widthForHeightFactor = (
            1.0 * pixmap.size().width() / pixmap.size().height()
        )
        self.setBackgroundBrush(QtGui.QBrush(pixmap))

    def hide_background(self):
        if self.background_on:
            self.background_on = False
            self.invalidate(self.sceneRect(), QtWidgets.QGraphicsScene.BackgroundLayer)

    def show_background(self):
        if not self.background_on:
            self.background_on = True
            self.invalidate(self.sceneRect(), QtWidgets.QGraphicsScene.BackgroundLayer)

    def mouseMoveEvent(self, event):
        self.fake_cursor.hide()
        # item = self.activePanel()
        # item.mouseMoveEvent(event)
        super().mouseMoveEvent(event)

    def mousePressEvent(self, event: "QGraphicsSceneMouseEvent") -> None:
        # Do not handle the event in ScrollHandDrag Mode to not interfere
        if self.views()[0].dragMode() == QtWidgets.QGraphicsView.ScrollHandDrag:
            return
        if event.button() == Qt.RightButton:
            self.activePanel().setFocus(Qt.OtherFocusReason)
            self.clearSelection()
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: "QGraphicsSceneMouseEvent") -> None:
        if self.views()[0].dragMode() == QtWidgets.QGraphicsView.ScrollHandDrag:
            return
        super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event: QtGui.QMouseEvent) -> None:
        item = self.activePanel()
        item.mouseDoubleClickEvent(event)
        super().mouseDoubleClickEvent(event)

    def contextMenuEvent(self, event: QGraphicsSceneContextMenuEvent) -> None:
        # item = self.activePanel()
        # item.contextMenuEvent(event)
        super().contextMenuEvent(event)



from PySide6 import QtCore, QtGui, QtWidgets

from eyelab.views.ui.ui_spline_options import Ui_splineOptions


class SplineWidget(QtWidgets.QWidget, Ui_splineOptions):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)

        # self.showCheckBox.setChecked()
        self.showCheckBox.stateChanged.connect(self.change_controlpoints)

        # Todo: fix bugs to enable features
        self.label.hide()
        self.slopeCheckBox.hide()
        self.strengthCheckBox.hide()
        self.neighbourCheckBox.hide()
        self.showCheckBox.hide()

    def change_controlpoints(self):
        scenetab = self.parentWidget().parentWidget()
        model_index = scenetab.imageTreeView.selectionModel().currentIndex()
        item = scenetab.model.getItem(model_index)
        if self.showCheckBox.isChecked():
            item.show_control_points()
        else:
            item.hide_control_points()

    def set_data(self, data: dict):
        checkbox_options = [
            (self.slopeCheckBox, "spline:slope"),
            (self.strengthCheckBox, "spline:strength"),
            (self.neighbourCheckBox, "spline:neighbour"),
        ]
        for cb, option in checkbox_options:
            if not option in data:
                data[option] = True
            cb.setChecked(data[option])
            cb.stateChanged.connect(lambda state: data.update({option: bool(state)}))


class PaintPreview(QtWidgets.QGraphicsItem):
    def __init__(self, settings_widget, parent=None):
        super().__init__(parent)
        self.settings_widget = settings_widget
        self.init_preview()

    def init_preview(self):
        pass

    def boundingRect(self) -> QtCore.QRectF:
        return QtCore.QRectF(0.0, 0.0, 0.0, 0.0)

    def paint(
        self, painter: QtGui.QPainter, option: "QStyleOptionGraphicsItem", widget
    ) -> None:
        return None


class Spline(object):
    def __init__(self):
        """ """
        self.name = "spline"
        self.cursor = self.get_cursor()
        self.button = self.get_tool_button()
        self.hot_key = None
        self.options_widget = self.get_options_widget()
        self.paint_preview = PaintPreview(self.options_widget)

    def get_options_widget(self):
        widget = SplineWidget()
        return widget

    def enable(self):
        pass

    def disable(self):
        pass

    def get_tool_button(self):
        button = QtWidgets.QToolButton()
        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(":/icons/icons/path-tool.svg"),
            QtGui.QIcon.Normal,
            QtGui.QIcon.Off,
        )
        button.setIcon(icon)
        button.setIconSize(QtCore.QSize(24, 24))
        button.setCheckable(True)
        button.setObjectName("inspectionButton")
        button.setToolTip("Curve Tool")
        return button

    def get_cursor(self):
        return QtGui.QCursor(
            QtGui.QPixmap(":/cursors/cursors/path_cursor.svg"), hotX=0, hotY=0
        )

    def mouse_move_handler(self, gitem, event):
        pass

    def mouse_press_handler(self, gitem, event):
        pass

    def mouse_doubleclick_handler(self, gitem, event):
        pos = event.scenePos()
        gitem.add_knot(pos)

    def mouse_release_handler(self, gitem, event):
        pass

    def mouse_release_handler(self, gitem, event):
        pass

    def key_press_handler(self, gitem, event):
        pass

    def key_release_handler(self, gitem, event):
        pass



import logging
from itertools import groupby
from typing import Any, Dict, List, Tuple

import eyepy as ep
import numpy as np
from PySide6.QtCore import QLineF, QPoint, QPointF, QRectF, Qt
from PySide6.QtGui import (
    QAction,
    QColor,
    QFocusEvent,
    QKeyEvent,
    QPainterPath,
    QPen,
    QPolygonF,
)
from PySide6.QtWidgets import (
    QApplication,
    QGraphicsEllipseItem,
    QGraphicsItem,
    QGraphicsLineItem,
    QGraphicsPathItem,
    QGraphicsRectItem,
    QGraphicsSceneContextMenuEvent,
    QGraphicsSceneHoverEvent,
    QGraphicsSceneMouseEvent,
    QMenu,
)
from scipy.interpolate import interp1d

from eyelab.commands import get_undo_stack
from eyelab.commands import layeritem as layer_commands
from eyelab.commands.layeritem import (
    DeleteCurve,
    DeletePolygon,
    MoveControlKnot,
    MoveKnot,
)
from eyelab.models.treeview.itemgroup import ItemGroup

logger = logging.getLogger(__name__)


class ControllPointGraphicsItem(QGraphicsRectItem):
    def __init__(self, parent, pos, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.setFlag(QGraphicsItem.ItemIgnoresTransformations, True)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        self.setFlag(QGraphicsItem.ItemSendsScenePositionChanges)

        self.setRect(QRectF(QPoint(-4, -4), QPoint(4, 4)))
        self.setPos(pos)

        pen = QPen(QColor("blue"))
        pen.setCosmetic(True)
        self.setPen(pen)

    @property
    def center(self):
        return self.mapToScene(QPointF(0, 0))

    def as_tuple(self):
        center = self.center
        return np.round(center.x(), 2), np.round(center.y(), 2)

    def mouseMoveEvent(self, event: "QGraphicsSceneMouseEvent") -> None:
        command = MoveControlKnot(self, self.mapToParent(event.pos()))
        get_undo_stack("main").push(command)
        super().mouseMoveEvent(event)

    def itemChange(self, change: QGraphicsItem.GraphicsItemChange, value: Any) -> Any:
        if (
            change == QGraphicsItem.ItemPositionChange
            or change == QGraphicsItem.ItemScenePositionHasChanged
        ):
            self.parentItem().set_line_of(self)

            if self is self.parentItem().cp_in:
                self.parentItem().knot_dict["cp_in_pos"] = self.as_tuple()
            elif self is self.parentItem().cp_out:
                self.parentItem().knot_dict["cp_out_pos"] = self.as_tuple()
            return value

        return value


class CubicSplineKnotItem(QGraphicsEllipseItem):
    def __init__(
        self,
        parent,
        knot_dict: dict,
        **kwargs,
    ):
        """"""
        super().__init__(parent=parent, **kwargs)
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        self.setFlag(QGraphicsItem.ItemIsFocusable, True)
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
        self.setFlag(QGraphicsItem.ItemIgnoresTransformations, True)

        self.knot_dict = knot_dict
        # Create knot
        self._cp_in = None
        self._cp_out = None

        self.setRect(QRectF(QPoint(-5, -5), QPoint(5, 5)))
        self.setPos(self.mapFromScene(self.knot_pos))
        pen = QPen(QColor("red"))
        pen.setCosmetic(True)
        self.setPen(pen)

        # Create control points
        pen = QPen(QColor("blue"))
        pen.setCosmetic(True)
        self.cps_visible = True

        self._cp_in = ControllPointGraphicsItem(self, self.mapFromScene(self.cp_in_pos))
        self._line_in = QGraphicsLineItem(
            QLineF(
                self.mapFromScene(self.center),
                self.mapFromScene(self.cp_in.center),
            ),
            parent=self,
        )
        self._line_in.setPen(pen)

        self._cp_out = ControllPointGraphicsItem(
            self, self.mapFromScene(self.cp_out_pos)
        )
        self._line_out = QGraphicsLineItem(
            QLineF(
                self.mapFromScene(self.center),
                self.mapFromScene(self.cp_out.center),
            ),
            parent=self,
        )
        self._line_out.setPen(pen)

    def itemChange(self, change: QGraphicsItem.GraphicsItemChange, value: Any) -> Any:
        if change == QGraphicsItem.ItemPositionChange and self.scene():
            newPos = QPointF(value)
            rect = self.scene().sceneRect()

            if not rect.contains(newPos):
                # Keep the item inside the scene rect.
                newPos.setX(min(rect.right(), max(newPos.x(), rect.left())))
                newPos.setY(min(rect.bottom(), max(newPos.y(), rect.top())))

            self.knot_dict["knot_pos"] = self.as_tuple()
            return newPos
        return value

    @property
    def bspline(self) -> "CubicSpline":
        return self.parentItem()

    @property
    def layer_item(self) -> "LayerItem":
        return self.parentItem().layer_item

    def as_tuple(self):
        center = self.center
        return np.round(center.x(), 2), np.round(center.y(), 2)

    def sync(self):
        self.knot_pos = self.as_tuple()
        if self.cp_in:
            self.cp_in_pos = self.cp_in.as_tuple()
        if self.cp_out:
            self.cp_out_pos = self.cp_out.as_tuple()

    @property
    def knot_pos(self):
        return QPointF(*self.knot_dict["knot_pos"])

    @knot_pos.setter
    def knot_pos(self, value: tuple):
        self.knot_dict["knot_pos"] = value

    @property
    def cp_in_pos(self):
        return QPointF(*self.knot_dict["cp_in_pos"])

    @cp_in_pos.setter
    def cp_in_pos(self, value: tuple):
        self.knot_dict["cp_in_pos"] = value

    @property
    def cp_out_pos(self):
        return QPointF(*self.knot_dict["cp_out_pos"])

    @cp_out_pos.setter
    def cp_out_pos(self, value: tuple):
        self.knot_dict["cp_out_pos"] = value

    def set_cp_out_length(self, length):
        line = QLineF(self.center, self.cp_out.center)
        line.setLength(length)
        self.cp_out = line.p2()

    def set_cp_in_length(self, length):
        line = QLineF(self.center, self.cp_in.center)
        line.setLength(length)
        self.cp_in = line.p2()

    @property
    def center(self) -> QPointF:
        return self.mapToScene(QPointF(0, 0))

    def hide_control_points(self):
        self.cps_visible = False
        self._cp_in.hide()
        self._cp_out.hide()
        self._line_in.hide()
        self._line_out.hide()

    def show_control_points(self):
        self.cps_visible = True
        self._cp_in.show()
        self._cp_out.show()
        self._line_in.show()
        self._line_out.show()

    @property
    def cp_in(self) -> ControllPointGraphicsItem:
        return self._cp_in

    @cp_in.setter
    def cp_in(self, cp):
        self._cp_in.setPos(self.mapFromScene(cp))
        self._set_line_in()
        self.cp_in_pos = self.cp_in.as_tuple()

    @property
    def cp_out(self) -> ControllPointGraphicsItem:
        return self._cp_out

    @cp_out.setter
    def cp_out(self, cp):
        self.cp_out.setPos(self.mapFromScene(cp))
        self._set_line_out()
        self.cp_out_pos = self.cp_out.as_tuple()

    def set_lines(self):
        if self.cps_visible:
            self._set_line_out()
            self._set_line_in()

    def set_line_of(self, cp: ControllPointGraphicsItem):
        if cp == self.cp_in:
            self._set_line_in()
        elif cp == self.cp_out:
            self._set_line_out()

    def _set_line_in(self):
        self._line_in.setLine(
            QLineF(self.mapFromScene(self.center), self.mapFromScene(self.cp_in.center))
        )

    def _set_line_out(self):
        self._line_out.setLine(
            QLineF(
                self.mapFromScene(self.center), self.mapFromScene(self.cp_out.center)
            )
        )

    def focusInEvent(self, event: QFocusEvent) -> None:
        pen = self.pen()
        pen.setStyle(Qt.DotLine)
        self.setPen(pen)
        super().focusInEvent(event)

    def focusOutEvent(self, event: QFocusEvent) -> None:
        pen = self.pen()
        pen.setStyle(Qt.SolidLine)
        self.setPen(pen)
        super().focusOutEvent(event)

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        if event.buttons() & Qt.RightButton:
            command = layer_commands.DeleteKnot(self)
            get_undo_stack("main").push(command)
            event.accept()
            return
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        index = self.bspline.knots.index(self)
        left_knot = self.bspline.knots[index - 1].center.x() if index > 0 else 0
        right_knot = (
            self.bspline.knots[index + 1].center.x()
            if index + 1 < len(self.bspline.knots)
            else self.scene().shape[1]
        )

        index = self.layer_item.cubic_splines.index(self.bspline)
        if index > 0 and left_knot == 0:
            left_knot = self.layer_item.cubic_splines[index - 1].knots[-1].center.x()
        if (
            index + 1 < len(self.layer_item.cubic_splines)
            and right_knot == self.scene().shape[1]
        ):
            right_knot = self.layer_item.cubic_splines[index + 1].knots[0].center.x()

        if left_knot < event.scenePos().x() < right_knot:
            command = MoveKnot(self, event.scenePos())
            get_undo_stack("main").push(command)
        event.accept()
        # super().mouseMoveEvent(event)

    def keyPressEvent(self, event: QKeyEvent) -> None:
        # modifiers = QApplication.keyboardModifiers()
        if event.key() == Qt.Key_Delete:
            spline = self.parentItem()
            knot = self
            command = layer_commands.DeleteKnot(knot)
            get_undo_stack("main").push(command)
            event.accept()
            return
        super().keyPressEvent(event)


class CubicSpline(QGraphicsPathItem):
    def __init__(self, knots: List[dict], parent: "LayerItem"):
        super().__init__(parent=parent)
        self._knots = knots
        self.control_points_visible = False
        # Add knots and controll points for every Curve
        self.knots = [
            CubicSplineKnotItem(parent=self, knot_dict=k) for k in self._knots
        ]

        self.hide_control_points()
        self.hide_knots()

        self.setAcceptHoverEvents(True)

        self.setFlag(QGraphicsItem.ItemIsFocusable)
        self.setFlag(QGraphicsItem.ItemIsSelectable)
        self.update()

    @property
    def start(self) -> QPointF:
        return self.knots[0].center

    @property
    def end(self) -> QPointF:
        return self.knots[-1].center

    @property
    def x_region(self) -> Tuple[float, float]:
        if len(self.knots) == 0:
            return 0.0, 0.0
        return self.start.x(), self.end.x()

    def __contains__(self, item):
        if self.start.x() < item.x() < self.end.x():
            return True
        return False

    def _build_path(self) -> QPainterPath:
        knots = self.knots
        if len(knots) == 0:
            return QPainterPath()

        path = QPainterPath(knots[0].center)
        last_knot = knots[0]
        if len(knots) == 1:
            return path

        for i, current_knot in enumerate(knots[1:], 1):
            current = current_knot.center
            path.cubicTo(last_knot.cp_out.center, current_knot.cp_in.center, current)
            last_knot = current_knot
        return path

    def hide_knots(self):
        for knot in self.knots:
            knot.hide()

    def show_knots(self):
        for knot in self.knots:
            knot.show()

    def hide_control_points(self):
        for knot in self.knots:
            knot.hide_control_points()
        self.control_points_visible = False

    def show_control_points(self):
        for knot in self.knots:
            knot.show_control_points()
        self.control_points_visible = True

    def indices(self):
        self.update()
        start = np.floor(self.start.x()).astype(int)
        end = np.floor(self.end.x()).astype(int)
        if start == end:
            return {start: self.start.y()}

        n_points = end - start + 1
        x_indices = []
        y_indices = []
        for t in np.linspace(0, 1, n_points * 2):
            point = self.path().pointAtPercent(t)
            x_indices.append(point.x())
            y_indices.append(point.y())

        x_indices = np.round(np.array(x_indices), 5)
        y_indices = np.array(y_indices)

        f = interp1d(
            x_indices, y_indices, assume_sorted=True, copy=False, bounds_error=True
        )
        x = np.arange(start, end + 1, dtype=int)
        # Get height for the middle of each pixel
        y = f(x + 0.5)

        return {xi: yi for xi, yi in zip(x, y)}

    def update(self):
        self.setPath(self._build_path())
        super().update()

    def add_knot(self, knot: CubicSplineKnotItem) -> None:
        knot.setParentItem(self)
        if len(self._knots) > 0:
            # Get insertion index
            i = 0
            for k in self._knots:
                if knot.center.x() > k["knot_pos"][0]:
                    i += 1
                    continue
                break
            self._knots.insert(i, knot.knot_dict)
            self.knots.insert(i, knot)
        else:
            self._knots.append(knot.knot_dict)
            self.knots.append(knot)

        if not self.control_points_visible:
            knot.hide_control_points()

    def shape(self) -> QPainterPath:
        # Create a path which closes without increasing its "area"
        # Only clicking exactly the line should activate this item
        path = self.path()
        path.connectPath(self.path().toReversed())
        return path

    def optimize_controllpoints(self, knot, distance_factor=0.35):
        knots = self.knots
        index = knots.index(knot)

        # options_widget = self.scene().current_tool.options_widget
        # optimize_strength = options_widget.strengthCheckBox.isChecked()
        # optimize_angle = options_widget.slopeCheckBox.isChecked()
        # optimize_neighbours = options_widget.neighbourCheckBox.isChecked()

        # meta = self.annotation_data.meta
        optimize_strength = True  # meta["spline:optimize_strength"]
        optimize_angle = True  # meta["spline:optimize_angle"]

        if len(knots) == 1:
            pos = knot.center
            source = QLineF(QPointF(pos.x() - 10, pos.y()), pos)
            target = QLineF(pos, QPointF(pos.x() + 10, pos.y()))

        elif index == 0:
            target = QLineF(knot.center, knots[1].center)
            source = (
                QLineF()
                .fromPolar(target.length(), 180 + target.angle())
                .translated(knot.center)
            )
            source.setPoints(source.p2(), source.p1())

        elif index == len(knots) - 1:
            source = QLineF(knots[-2].center, knot.center)
            target = (
                QLineF()
                .fromPolar(source.length(), source.angle())
                .translated(knot.center)
            )
        else:
            source = QLineF(knots[index - 1].center, knot.center)
            target = QLineF(knot.center, knots[index + 1].center)

        targetAngle = target.angleTo(source)
        if targetAngle > 180:
            angle = (source.angle() + source.angleTo(target) / 2) % 360
        else:
            angle = (target.angle() + target.angleTo(source) / 2) % 360

        if optimize_strength:
            length_in = source.dx() * distance_factor
            length_out = target.dx() * distance_factor
        else:
            length_in = knot._line_in.line().length()
            length_out = knot._line_out.line().length()

        if optimize_angle:
            angle_in = angle + 180
            angle_out = angle
        else:
            angle_in = (knot._line_in.line().angle()) % 360
            angle_out = knot._line_out.line().angle()

        revTarget = QLineF.fromPolar(length_in, angle_in).translated(knot.center)
        cp_in = revTarget.p2()

        revSource = QLineF.fromPolar(length_out, angle_out).translated(knot.center)
        cp_out = revSource.p2()
        return cp_in, cp_out

    @property
    def layer_item(self) -> "LayerItem":
        return self.parentItem()

    def hoverEnterEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        pen = self.pen()
        pen.setWidth(pen.width() + 3)
        self.setPen(pen)
        super().hoverEnterEvent(event)

    def hoverLeaveEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        pen = self.pen()
        pen.setWidth(pen.width() - 3)
        self.setPen(pen)
        super().hoverLeaveEvent(event)

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(event)

    def focusInEvent(self, event: QFocusEvent) -> None:
        pen = self.pen()
        pen.setStyle(Qt.DotLine)
        self.setPen(pen)
        super().focusInEvent(event)

    def focusOutEvent(self, event: QFocusEvent) -> None:
        pen = self.pen()
        pen.setStyle(Qt.SolidLine)
        self.setPen(pen)
        super().focusOutEvent(event)

    def keyPressEvent(self, event: QKeyEvent) -> None:
        if event.key() == Qt.Key_Delete:
            command = DeleteCurve(self)
            get_undo_stack("main").push(command)
            event.accept()
            return
        super().keyPressEvent(event)


class PolygonPath(QGraphicsPathItem):
    def __init__(self, parent, heights, start: QPointF = None, end: QPointF = None):
        super().__init__(parent)
        self.heights = heights
        self._start = start
        self._end = end

        # self.setAcceptHoverEvents(True)
        # self.setFlag(QGraphicsItem.ItemIsFocusable)
        self.update()

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        # print(self.layer_item.focusItem())
        super().mousePressEvent(event)
        # print(self.layer_item.focusItem())

    def shape(self) -> QPainterPath:
        # Create a path which closes without increasing its "area"
        # Only clicking exactly the line should activate this item
        path = self.path()
        path.connectPath(self.path().toReversed())
        return path

    @property
    def layer_item(self) -> "LayerItem":
        return self.parentItem()

    def _get_points(self):
        points = []
        if self._start is None:
            start_index = 0
        else:
            start_index = int(np.floor(self._start.x()))
            points.append(self._start)

        if self._end is None:
            end_index = len(self.heights)
        else:
            end_index = int(np.ceil(self._end.x()))

        # Make sure there are no Nans between start and stop - interpolate
        if any(np.isnan(self.heights[start_index:end_index])):
            x = np.arange(start_index, end_index, dtype=int)
            x_full = x[~np.isnan(self.heights[x])]
            x_empty = x[np.isnan(self.heights[x])]
            if len(x_full) > 1:
                f = interp1d(
                    x_full,
                    self.heights[x_full],
                    assume_sorted=True,
                    copy=False,
                    bounds_error=False,
                    fill_value="extrapolate",
                )
                # Get height for the middle of each pixel
                y = f(x_empty + 0.5)
                self.heights[x_empty] = y

        for i in range(start_index, end_index):
            # +0.5 for points to sit in the middle and not start of each pixel (in x direction)
            points.append(QPointF(i + 0.5, self.heights[i]))

        if self._end is not None:
            points.append(self._end)
        return points

    def update(self):
        points = self._get_points()
        if points == []:
            self.setPath(QPainterPath())
            super().update()
            return

        self.polygon = QPolygonF().fromList(points)
        path = QPainterPath()
        path.addPolygon(self.polygon)
        self.setPath(path)
        super().update()

    @property
    def start(self) -> QPointF:
        return self._start

    @start.setter
    def start(self, point: QPointF) -> None:
        self._start = point
        self.update()

    @property
    def end(self) -> QPointF:
        return self._end

    @end.setter
    def end(self, point: QPointF) -> None:
        self._end = point
        self.update()

    @property
    def x_region(self) -> Tuple[float, float]:
        return self.start.x(), self.end.x()

    def __contains__(self, item):
        if self.start.x() < item.x() < self.end.x():
            return True
        return False

    def hoverEnterEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        pen = self.pen()
        pen.setWidth(pen.width() + 3)
        self.setPen(pen)
        super().hoverEnterEvent(event)

    def hoverLeaveEvent(self, event: QGraphicsSceneHoverEvent) -> None:
        pen = self.pen()
        pen.setWidth(pen.width() - 3)
        self.setPen(pen)
        super().hoverLeaveEvent(event)

    def focusInEvent(self, event: QFocusEvent) -> None:
        pen = self.pen()
        pen.setStyle(Qt.DotLine)
        self.setPen(pen)
        super().focusInEvent(event)

    def focusOutEvent(self, event: QFocusEvent) -> None:
        pen = self.pen()
        pen.setStyle(Qt.SolidLine)
        self.setPen(pen)
        super().focusOutEvent(event)

    def keyPressEvent(self, event: QKeyEvent) -> None:
        if event.key() == Qt.Key_Delete:
            command = DeletePolygon(self)
            get_undo_stack("main").push(command)
            event.accept()
            return
        super().keyPressEvent(event)


class LayerItem(QGraphicsPathItem):
    def __init__(
        self, data: ep.EyeVolumeLayerAnnotation, index: int, parent: ItemGroup
    ):
        super().__init__(parent=parent)
        self.annotation_data = data
        self.index = index

        # Make sure knots are List[Curve[KnotDict]] and not List[KnotDict]
        knots = self.annotation_data.knots[self.index]
        if knots:
            if type(knots[0]) == dict:
                self.annotation_data.knots[self.index] = [knots]
        self.cubic_splines = [CubicSpline(knots, self) for knots in self.knots]
        self.polygons = self._get_polygons()

        self.setFlag(QGraphicsItem.ItemIsPanel)
        self.setFlag(QGraphicsItem.ItemIsFocusable)
        self.setFlag(QGraphicsItem.ItemIsSelectable)

        for cs in self.cubic_splines:
            left, right = self.get_neighbour_elements(cs)
            if left:
                left.end = cs.start
            if right:
                right.start = cs.end
        self.update()

    def get_neighbour_elements(self, layer_element):
        elements = sorted(self.cubic_splines + self.polygons, key=lambda x: x.start.x())
        index = elements.index(layer_element)
        left = elements[index - 1] if index > 0 else None
        right = elements[index + 1] if index + 1 < len(elements) else None
        return left, right

    def get_neighbour_polygons(self, layer_element):
        """Get polygons starting left and right from the given element

        The returned elements are ordered by their distance to the given elements

        :returns (left_neighbours, right_neighbours)
        """
        elements = sorted(
            self.polygons + [layer_element], key=lambda x: (x.start.x(), x.end.x())
        )
        index = elements.index(layer_element)
        return elements[:index][::-1], elements[index + 1 :]

    @property
    def knots(self):
        return self.annotation_data.knots[self.index]

    def _get_polygons(self) -> List[PolygonPath]:
        """Create polygons from the layer height maps in regions outside the cubic splines

        Polygons cover A-scans that are not part of a Cubic Spline and are connected to the first/last knot
        of neighbouring Cubic splines.
        """
        # Get polygon regions:
        layer_copy = self.height_map.copy()
        # Create a polygon for every non spline region
        for cs in self.cubic_splines:
            layer_copy[int(np.floor(cs.start.x())) : int(np.ceil(cs.end.x()))] = np.nan

        layer_copy = ~np.isnan(layer_copy)
        # Collect ranges of polygon regions (python indexing)
        i = 0
        polygon_regions = []

        for k, g in groupby(layer_copy):
            l = len(list(g))
            if k:
                polygon_regions.append(
                    (
                        QPointF(i + 0.5, self.height_map[i]),
                        QPointF(i + l - 0.5, self.height_map[i + l - 1]),
                    )
                )
            i += l

        return [self._get_polygon(start, end) for start, end in polygon_regions]

    def _get_polygon(self, start: QPointF, end: QPointF) -> PolygonPath:
        return PolygonPath(self, self.height_map, start, end)

    @property
    def height_map(self):
        return self.annotation_data.data[-(self.index + 1)]

    def setActive(self, active: bool) -> None:
        if active:
            for cs in self.cubic_splines:
                cs.show_knots()
                # cs.show_control_points()
        else:
            for cs in self.cubic_splines:
                cs.hide_knots()
                # cs.hide_control_points()
        super().setActive(active)

    def as_array(self):
        """Return the annotated path as an array of shape (image width)

        The array has the same shape as the annotated image width. Regions
        which are not annotated become np.nan

        The array is build by painting the annotated path on a pixmap,
        converting it to a numpy array and computing the column-wise center of
        mass for the first channel.
        """
        heights = self.annotation_data.data[-(self.index + 1)]

        spline_region = self.cubic_spline.x_region
        width = spline_region[1] - spline_region[0]

        for t in np.linspace(0, 1, np.ceil(width).astype(int), endpoint=False):
            path = self.path()
            point = path.pointAtPercent(t)
            x = np.rint(point.x() - 0.5).astype(int)
            heights[x] = point.y() - 0.5

        return heights

    def update(self):
        self.setVisible(self.annotation_data.meta["visible"])
        self.setZValue(self.annotation_data.meta["z_value"])

        color = QColor()
        color.setNamedColor(f"#{self.annotation_data.meta['current_color']}")

        pen = QPen(color)
        pen.setWidth(2)
        pen.setCosmetic(True)
        self.setPen(pen)

        for cs in self.cubic_splines:
            cs.setPen(pen)
            cs.update()

        for p in self.polygons:
            p.setPen(pen)
            p.update()

        # x, y = self.cubic_spline.indices
        # self.annotation_data.data[-(self.index + 1), x] = y
        super().update()

    def view(self):
        return self.scene().views()[0]

    def mouseDoubleClickEvent(self, event):
        modifiers = QApplication.keyboardModifiers()
        if modifiers == Qt.ControlModifier:
            print(self.polygons, self.cubic_splines)
            event.ignore()
            return
        # if event.modifiers() &
        pos = event.scenePos()
        # Ignore event if not inside the scene
        if not self.scene().sceneRect().contains(pos):
            event.ignore()
            return

        # Find polygon/curve intersecting the new Knot
        cs = [cs for cs in self.cubic_splines if pos in cs]
        # if pos is in any existing curve add the knot to this curve
        if cs:
            command = layer_commands.AddKnot(cs[0], pos)
        elif type(self.focusItem()) == LayerItem or self.focusItem() is None:
            command = layer_commands.AddCurve(self, event.scenePos())
        # else if there is a focusitem add the knot to this items curve
        else:
            if type(self.focusItem()) is CubicSpline:
                bspline = self.focusItem()
            elif type(self.focusItem()) is CubicSplineKnotItem:
                bspline = self.focusItem().bspline
            else:
                print("unexpected focus item ", self.focusItem())
                return
            # Ignore event if there is other curve in between focusItem and new knot
            if any(
                [
                    pos.x() < cs.start.x() < bspline.start.x()
                    or bspline.end.x() < cs.end.x() < pos.x()
                    for cs in self.cubic_splines
                ]
            ):
                return
            command = layer_commands.AddKnot(bspline, pos)

        get_undo_stack("main").push(command)
        super().mouseDoubleClickEvent(event)

    def childNumber(self):
        if self.parentItem():
            return self.parentItem().childItems().index(self)
        return 0

    def childCount(self):
        return 0

    def columnCount(self):
        return 1

    def data(self, column: str):
        if column in ["visible", "z_value", "current_color"]:
            return getattr(self, column)
        elif column == "name":
            return self.meta["name"]

        raise Exception(f"column {column} not in data")

    def setData(self, column: str, value):
        if column in ["visible", "z_value", "current_color"]:
            setattr(self, column, value)
            self.update()
            return True
        return False

    def appendChild(self, data: "TreeLineItem"):
        items = self.childItems()

        if items:
            z_value = float(items[-1].zValue() + 1)
        else:
            z_value = 0.0

        data.z_value = z_value
        data.setParentItem(self)

    def insertChildren(self, row: int, count: int, data: List[Dict] = None):
        if row < 0:
            return False

        items = self.childItems()

        if items:
            z = float(items[-1].zValue() + 1)
        else:
            z = 0.0
        z_values = [float(x) for x in range(z, z + count)]

        for i, z_value in enumerate(z_values):
            if data:
                item_data = data[i]
            else:
                item_data = {}
            item_data.update(z_value=z_value)
            layer = type(self)(data=item_data)
            layer.setParentItem(self)

    def removeChildren(self, row: int, count: int):
        if row < 0 or row > self.childCount():
            raise Exception("what went wrong here?")
        items = self.childItems()

        for i in range(row, row + count):
            item = items[i]
            item.scene().removeItem(item)

    def switchChildren(self, row1: int, row2: int):
        child1 = self.child(row1)
        child2 = self.child(row2)

        child1_z = child1.zValue()
        child2_z = child2.zValue()
        child1.setData("z_value", child2_z)
        child2.setData("z_value", child1_z)

    # def contextMenuEvent(self, event: QGraphicsSceneContextMenuEvent) -> None:
    #    menu = QMenu()
    #    new_curve_action = QAction("New Curve")
    #    menu.addAction(new_curve_action)

    #    menu.exec(event.screenPos())

    #    new_curve_action.triggered.connect(lambda: self.setFocus())
    #    #super().contextMenuEvent(event)



from eyelab.dialogs.help.help_dialog import HelpWindow

help_text = """
# Shortcut Sheet

## Annotation

+ Ctrl + X: toggle linked navigation
"""


class ShortcutHelp(HelpWindow):
    def __init__(self, parent):
        super().__init__(help_text, parent)



import logging

import numpy as np
from eyepy import EyeVolume
from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import QPointF

from eyelab.models.scene import Line, Point
from eyelab.models.viewtab import VolumeTab
from eyelab.views.graphicsview import CustomGraphicsView

logger = logging.getLogger("eyelab.volumeview")


class VolumeView(CustomGraphicsView):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.data = None
        self._lines = None

        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)

    def reset(self):
        self.data = None
        self._lines = None

    def set_data(self, data: EyeVolume, name: str = "Volume"):
        logger.debug("VolumeView: set_data")
        self.reset()
        self.data = data
        self.name = name
        self.view_tab = VolumeTab(self.data, parent=self)
        self.setScene(self.view_tab.model.scene)

        self.zoomToFit()
        logger.debug("VolumeView: data is set")

    def map_to_localizer(self, pos):
        slice_n = pos.toPoint().y()
        lclzr_scale_x = self.data.localizer.scale_x
        lclzr_scale_y = self.data.localizer.scale_y
        start_y = self.data[slice_n].meta["start_pos"][1] / lclzr_scale_y
        end_y = self.data[slice_n].meta["end_pos"][1] / lclzr_scale_y
        size_x = self.data.size_x

        x = self.data[slice_n].meta["start_pos"][0] / lclzr_scale_x + pos.x()
        y = start_y + (start_y - end_y) / size_x * pos.x()

        return QPointF(x, y)

    def map_from_localizer(self, pos):
        lclzr_scale_x = self.data.localizer.scale_x
        current_slice = self.view_tab.model.current_slice
        x = pos.x() - self.data[current_slice].meta["start_pos"][0] / lclzr_scale_x
        y = self.closest_slice(pos)
        return QPointF(x, y)

    def set_fake_cursor(self, pos, sender):
        # Turn localizer position to x pos and slice number for OCT
        pos = self.map_from_localizer(pos)
        # set slice

        if self.linked_navigation:
            self.view_tab.model.current_slice = int(pos.y())
            self.setScene(self.view_tab.model.scene)
            self.centerOn(pos)

        current_center = self.mapToScene(self.rect().center()).y()
        pos = QPointF(pos.x(), current_center)

        self.scene().fake_cursor.setPos(pos)
        self.scene().fake_cursor.show()

    def wheelEvent(self, event):
        if event.modifiers() == (QtCore.Qt.ControlModifier):
            if event.angleDelta().y() > 0:
                self.view_tab.next_slice()
            else:
                self.view_tab.last_slice()
            self.setScene(self.view_tab.model.scene)
            self.view_tab.model.activate(
                self.view_tab.imageTreeView.selectionModel().currentIndex()
            )

            if self.linked_navigation:
                current_slice = self.view_tab.model.current_slice
                pos_on_localizer = self.map_to_localizer(
                    QPointF(
                        self.mapToScene(event.position().toPoint()).x(), current_slice
                    )
                )
                self.cursorPosChanged.emit(pos_on_localizer, self)
            event.accept()
        else:

            super().wheelEvent(event)

    def mouseMoveEvent(self, event):
        super().mouseMoveEvent(event)
        scene_pos = self.mapToScene(event.pos())
        current_slice = self.view_tab.model.current_slice
        localizer_pos = self.map_to_localizer(
            QtCore.QPointF(scene_pos.x(), current_slice)
        )
        self.cursorPosChanged.emit(localizer_pos, self)

    def _slice_lines(self):
        lines = []
        for bscan in self.data:
            scale = np.array([self.data.localizer.scale_x, self.data.localizer.scale_y])

            start = bscan.meta["start_pos"] / scale
            end = bscan.meta["end_pos"] / scale

            p1 = Point(*start)
            p2 = Point(*end)
            a = p1.y - p2.y
            b = p2.x - p1.x
            c = a * p2.x + b * p2.y
            lines.append(Line(a, b, -c))
        return lines

    @property
    def slice_lines(self):
        if self._lines is None:
            self._lines = self._slice_lines()

        return self._lines

    def closest_slice(self, pos):
        # Todo: Make this faster for smooth registered navigation
        point = Point(pos.x(), pos.y())

        smallest_dist = self.point_line_distance(point, self.slice_lines[0])
        for i, line in enumerate(self.slice_lines):
            dist = self.point_line_distance(point, line)
            if dist <= smallest_dist:
                smallest_dist = dist
            else:
                return i - 1
        return i

    @staticmethod
    def point_line_distance(point, line):
        return np.abs(line.a * point.x + line.b * point.y + line.c) / np.sqrt(
            line.a ** 2 + line.b ** 2
        )



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_layergroup_entry.ui'
##
## Created by: Qt User Interface Compiler version 6.1.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *  # type: ignore
from PySide6.QtGui import *  # type: ignore
from PySide6.QtWidgets import *  # type: ignore

from . import resources_rc


class Ui_LayerGroupEntry(object):
    def setupUi(self, LayerGroupEntry):
        if not LayerGroupEntry.objectName():
            LayerGroupEntry.setObjectName(u"LayerGroupEntry")
        LayerGroupEntry.resize(200, 30)
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(LayerGroupEntry.sizePolicy().hasHeightForWidth())
        LayerGroupEntry.setSizePolicy(sizePolicy)
        LayerGroupEntry.setMinimumSize(QSize(200, 30))
        LayerGroupEntry.setMaximumSize(QSize(350, 30))
        font = QFont()
        font.setPointSize(10)
        LayerGroupEntry.setFont(font)
        LayerGroupEntry.setContextMenuPolicy(Qt.PreventContextMenu)
        LayerGroupEntry.setAutoFillBackground(True)
        self.horizontalLayout_2 = QHBoxLayout(LayerGroupEntry)
        self.horizontalLayout_2.setSpacing(2)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(2, 2, 2, 2)
        self.hideButton = QToolButton(LayerGroupEntry)
        self.hideButton.setObjectName(u"hideButton")
        sizePolicy1 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.hideButton.sizePolicy().hasHeightForWidth())
        self.hideButton.setSizePolicy(sizePolicy1)
        self.hideButton.setMinimumSize(QSize(26, 26))
        self.hideButton.setMaximumSize(QSize(26, 26))
        self.hideButton.setContextMenuPolicy(Qt.NoContextMenu)
        self.hideButton.setAutoFillBackground(False)
        icon = QIcon()
        icon.addFile(
            u":/icons/icons/baseline-visibility-24px.svg",
            QSize(),
            QIcon.Normal,
            QIcon.Off,
        )
        self.hideButton.setIcon(icon)
        self.hideButton.setIconSize(QSize(24, 24))

        self.horizontalLayout_2.addWidget(self.hideButton)

        self.label = QLabel(LayerGroupEntry)
        self.label.setObjectName(u"label")

        self.horizontalLayout_2.addWidget(self.label)

        self.retranslateUi(LayerGroupEntry)

        QMetaObject.connectSlotsByName(LayerGroupEntry)

    # setupUi

    def retranslateUi(self, LayerGroupEntry):
        LayerGroupEntry.setWindowTitle(
            QCoreApplication.translate("LayerGroupEntry", u"Form", None)
        )
        self.hideButton.setText(
            QCoreApplication.translate("LayerGroupEntry", u"...", None)
        )
        self.label.setText("")

    # retranslateUi



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_add_annotation_dialog.ui'
##
## Created by: Qt User Interface Compiler version 6.1.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *  # type: ignore
from PySide6.QtGui import *  # type: ignore
from PySide6.QtWidgets import *  # type: ignore

from . import resources_rc


class Ui_AddAnnotationDialog(object):
    def setupUi(self, AddAnnotationDialog):
        if not AddAnnotationDialog.objectName():
            AddAnnotationDialog.setObjectName(u"AddAnnotationDialog")
        AddAnnotationDialog.resize(400, 112)
        self.verticalLayout = QVBoxLayout(AddAnnotationDialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.formLayout = QFormLayout()
        self.formLayout.setObjectName(u"formLayout")
        self.nameLabel = QLabel(AddAnnotationDialog)
        self.nameLabel.setObjectName(u"nameLabel")
        font = QFont()
        font.setBold(True)
        self.nameLabel.setFont(font)

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.nameLabel)

        self.nameEdit = QLineEdit(AddAnnotationDialog)
        self.nameEdit.setObjectName(u"nameEdit")

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.nameEdit)

        self.typeLabel = QLabel(AddAnnotationDialog)
        self.typeLabel.setObjectName(u"typeLabel")
        self.typeLabel.setFont(font)

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.typeLabel)

        self.typeBox = QComboBox(AddAnnotationDialog)
        self.typeBox.setObjectName(u"typeBox")

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self.typeBox)

        self.verticalLayout.addLayout(self.formLayout)

        self.buttonBox = QDialogButtonBox(AddAnnotationDialog)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setEnabled(True)
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        self.buttonBox.setCenterButtons(False)

        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(AddAnnotationDialog)
        self.buttonBox.accepted.connect(AddAnnotationDialog.accept)
        self.buttonBox.rejected.connect(AddAnnotationDialog.reject)

        QMetaObject.connectSlotsByName(AddAnnotationDialog)

    # setupUi

    def retranslateUi(self, AddAnnotationDialog):
        AddAnnotationDialog.setWindowTitle(
            QCoreApplication.translate("AddAnnotationDialog", u"Dialog", None)
        )
        self.nameLabel.setText(
            QCoreApplication.translate("AddAnnotationDialog", u"Name:", None)
        )
        self.nameEdit.setText(
            QCoreApplication.translate("AddAnnotationDialog", u"New Annotation", None)
        )
        self.typeLabel.setText(
            QCoreApplication.translate("AddAnnotationDialog", u"Type:", None)
        )

    # retranslateUi



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_2d_viewer.ui'
##
## Created by: Qt User Interface Compiler version 6.1.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from oat.views import CustomGraphicsView
from PySide6.QtCore import *  # type: ignore
from PySide6.QtGui import *  # type: ignore
from PySide6.QtWidgets import *  # type: ignore


class Ui_Viewer2D(object):
    def setupUi(self, Viewer2D):
        if not Viewer2D.objectName():
            Viewer2D.setObjectName(u"Viewer2D")
        Viewer2D.resize(400, 400)
        Viewer2D.setMinimumSize(QSize(400, 400))
        self.verticalLayout = QVBoxLayout(Viewer2D)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.graphicsView2D = CustomGraphicsView(Viewer2D)
        self.graphicsView2D.setObjectName(u"graphicsView2D")
        self.graphicsView2D.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.graphicsView2D.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        self.verticalLayout.addWidget(self.graphicsView2D)

        self.retranslateUi(Viewer2D)

        QMetaObject.connectSlotsByName(Viewer2D)

    # setupUi

    def retranslateUi(self, Viewer2D):
        Viewer2D.setWindowTitle(QCoreApplication.translate("Viewer2D", u"Form", None))

    # retranslateUi



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eyelab',
 'eyelab.commands',
 'eyelab.dialogs',
 'eyelab.dialogs.help',
 'eyelab.models',
 'eyelab.models.treeview',
 'eyelab.tools',
 'eyelab.views',
 'eyelab.views.ui',
 'eyelab.views.ui.cursors',
 'eyelab.views.ui.icons']

package_data = \
{'': ['*']}

install_requires = \
['PySide6==6.1.3',
 'eyepie>=0.6.7,<0.7.0',
 'qimage2ndarray==1.9.0',
 'requests>=2.27.1,<3.0.0',
 'scikit-image>=0.19.1,<0.20.0']

setup_kwargs = {
    'name': 'eyelab',
    'version': '0.3.1',
    'description': 'Multi-modal annotation tool for eye imaging data',
    'long_description': None,
    'author': 'Olivier Morelle',
    'author_email': 'oli4morelle@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



import logging
import math

from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtWidgets import QGraphicsView

logger = logging.getLogger(__name__)


class CustomGraphicsView(QGraphicsView):
    cursorPosChanged = QtCore.Signal(QtCore.QPointF, QGraphicsView)

    viewChanged = QtCore.Signal(QGraphicsView)

    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self._zoom = 0

        self.linked_navigation = False

        # How to position the scene when transformed (eg zoom)
        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorViewCenter)
        # How to position the scene when resizing the widget
        self.setResizeAnchor(QtWidgets.QGraphicsView.AnchorViewCenter)
        # Get Move events even if no button is pressed
        self.setMouseTracking(True)
        self.deactivate_scroll_bars()

        self.setLineWidth(3)
        self.setPalette(QtGui.QPalette(QtCore.Qt.red, QtCore.Qt.black))
        self.setEnabled(False)

    def unlink_navigation(self):
        self.linked_navigation = False

    def link_navigation(self):
        self.linked_navigation = True

    @property
    def tool(self):
        return self.view_tab.current_tool

    def setEnabled(self, a0: bool) -> None:
        if a0:
            self.setFrameStyle(QtWidgets.QFrame.Raised | QtWidgets.QFrame.Panel)
        else:
            self.setFrameStyle(QtWidgets.QFrame.Plain | QtWidgets.QFrame.Panel)

        super().setEnabled(a0)

    def enterEvent(self, event):
        self.setEnabled(True)
        self.update_tool()

        if self.linked_navigation:
            tab_widget = self.view_tab.parent().parent()
            index = tab_widget.indexOf(self.view_tab)
            tab_widget.setCurrentIndex(index)

        self.scene().addItem(self.tool.paint_preview)

        super().enterEvent(event)

    def leaveEvent(self, event):
        self.setEnabled(False)
        if self.tool.paint_preview.scene() == self.scene():
            self.scene().removeItem(self.tool.paint_preview)
        super().leaveEvent(event)

    def mouseMoveEvent(self, event):
        super().mouseMoveEvent(event)
        scene_pos = self.mapToScene(event.pos())
        if self.tool.paint_preview.scene() == self.scene():
            self.tool.paint_preview.setPos(scene_pos.toPoint())
        self.cursorPosChanged.emit(scene_pos, self)

    def hasWidthForHeight(self):
        return True

    def widthForHeight(self, height):
        return math.ceil(height * self.scene()._widthForHeightFactor)

    def toggle_scroll_bars(self):
        if self._scroll_bars:
            self.deactivate_scroll_bars()
        else:
            self.activate_scroll_bars()

    def deactivate_scroll_bars(self):
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self._scroll_bars = False

    def activate_scroll_bars(self):
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self._scroll_bars = True

    def zoomToFit(self):
        self.fitInView(self.scene().sceneRect(), QtCore.Qt.KeepAspectRatio)
        self._zoom = 0

    def zoomToFeature(self):
        # Zoom in as long as more than 1/3 of width of the image is
        # visible
        while (
            self.mapToScene(self.rect()).boundingRect().width()
            > self.scene().width() / 3
        ):
            self.zoom_in()

    def zoom_in(self):
        self._zoom += 1
        self.scale(1.25, 1.25)

    def zoom_out(self):
        if self._zoom > 0:
            self._zoom -= 1
            self.scale(0.8, 0.8)

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        if event.angleDelta().y() > 0:
            self.zoom_in()
        else:
            self.zoom_out()

    def showEvent(self, event: QtGui.QShowEvent) -> None:
        super().showEvent(event)
        self.zoomToFit()

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        super().resizeEvent(event)
        self.zoomToFit()

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        super().mouseReleaseEvent(event)
        self.update_tool()

    def setScene(self, scene) -> None:
        super().setScene(scene)

    def update_tool(self):
        tool = self.view_tab.current_tool
        if tool.name == "inspection":
            self.setDragMode(QtWidgets.QGraphicsView.ScrollHandDrag)
        else:
            self.setDragMode(QtWidgets.QGraphicsView.NoDrag)
        self.viewport().setCursor(tool.cursor)



from eyelab.dialogs.help.help_dialog import HelpWindow

help_text = """
# Layer Annotation Guide

An OCT Layer in Eyelab is combination of explicit layer heights provided per A-scan,
and a cubic spline curve which can be manipulated by adding, removing or moving the
curves knots.

## Adding a new Layer

Create a new layer using the (+) symbol in the overview on the right. Specify a name
and select Layers from the Type dropdown.

## Layer Annotation

First select the layer you want to annotate from the overview on the right. Selecting a
layer automatically enables the **Curve** tool. You can now start adding control points
of a cubic spline curve, by double click on the desired location. You can move control
points by left click and hold and delete them with a right click.
These control points consist of a red circle the curve has to pass through and two
blue rectangles which control at which angle the curve enters and leaves the red control
point

If you want to pan around in the image select the **Inspection** tool on the right.

The layer heights are constantly updated to represent the position of the current cubic
spline curve.
"""


class LayerAnnotationHelp(HelpWindow):
    def __init__(self, parent):
        super().__init__(help_text, parent)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_layer_entry.ui'
##
## Created by: Qt User Interface Compiler version 6.1.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *  # type: ignore
from PySide6.QtGui import *  # type: ignore
from PySide6.QtWidgets import *  # type: ignore

from . import resources_rc


class Ui_LayerEntry(object):
    def setupUi(self, LayerEntry):
        if not LayerEntry.objectName():
            LayerEntry.setObjectName(u"LayerEntry")
        LayerEntry.resize(150, 30)
        LayerEntry.setMinimumSize(QSize(150, 30))
        LayerEntry.setMaximumSize(QSize(300, 30))
        LayerEntry.setContextMenuPolicy(Qt.ActionsContextMenu)
        LayerEntry.setAutoFillBackground(False)
        LayerEntry.setStyleSheet(u"background-color: white")
        self.horizontalLayout = QHBoxLayout(LayerEntry)
        self.horizontalLayout.setSpacing(2)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(2, 2, 2, 2)
        self.hideButton = QToolButton(LayerEntry)
        self.hideButton.setObjectName(u"hideButton")
        self.hideButton.setMinimumSize(QSize(26, 26))
        self.hideButton.setMaximumSize(QSize(26, 26))
        self.hideButton.setContextMenuPolicy(Qt.NoContextMenu)
        self.hideButton.setStyleSheet(u"background-color: white")
        icon = QIcon()
        icon.addFile(
            u":/icons/icons/baseline-visibility-24px.svg",
            QSize(),
            QIcon.Normal,
            QIcon.Off,
        )
        self.hideButton.setIcon(icon)
        self.hideButton.setIconSize(QSize(24, 24))

        self.horizontalLayout.addWidget(self.hideButton)

        self.colorButton = QToolButton(LayerEntry)
        self.colorButton.setObjectName(u"colorButton")
        self.colorButton.setMinimumSize(QSize(26, 26))
        self.colorButton.setContextMenuPolicy(Qt.NoContextMenu)
        self.colorButton.setAutoFillBackground(False)
        self.colorButton.setStyleSheet(u"background-color: white")

        self.horizontalLayout.addWidget(self.colorButton)

        self.label = QLabel(LayerEntry)
        self.label.setObjectName(u"label")

        self.horizontalLayout.addWidget(self.label)

        self.retranslateUi(LayerEntry)

        QMetaObject.connectSlotsByName(LayerEntry)

    # setupUi

    def retranslateUi(self, LayerEntry):
        LayerEntry.setWindowTitle(
            QCoreApplication.translate("LayerEntry", u"Form", None)
        )
        self.hideButton.setText(QCoreApplication.translate("LayerEntry", u"...", None))
        self.colorButton.setText("")
        self.label.setText("")

    # retranslateUi
